/*

Copyright (C) 1992,1993 by Roger Hirst

Version 1.1

ObjectCallBack is a class which allows Windows Call Back functions to call
non virtual class member functions. The technique used is similar to that
used by OWL. An instance thunk is built which includes the object's this
pointer.

Two interface functions are provided:

FARPROC WINAPI MakeProcInstance (CallObjTyp CallObjPtr,PCObject pThis)

FARPROC WINAPI FreeProcInstance (FARPROC lpfPtr);

These functions are designed to operate in much the same way as
the Windows functions MakeProcInstance and FreeProcInstance.

Up to 256 call back functions may be registered at any time in any
instance of ObjectCallBack.

*/

#include <owl.h>
#include "callback.h"

/*

ObjectCallBack::ObjectCallBack ();

Function:

The constructor for the class.

Description:

A 4Kb array of 16 byte thunks is allocated in global memory. This array
is initialised with the basic thunk code. The pointer details are filled
in by ObjectCallBack::MakeProcInstance.

PrestoChangoSelector (FP_SEG(vfpCode),FP_SEG(vfpCode)) changes the data
selector into a code selector so that the code written to a data area
can be executed without generating GP fails.

Returns:

Nothing

If anything fails the Status is set to FALSE.

*/

ObjectCallBack::ObjectCallBack ()
{
	BYTE code [InstanceCodeSize];

	Status = TRUE;
	pFreeList = new FreeListTyp ();
	if (pFreeList)
	{
		pCallList = new CallListTyp ();
		if (pCallList)
		{
			memset (&code,0,InstanceCodeSize);
#if defined(__HUGE__) || defined(__LARGE__)
			code [0] = POP_AX;
			code [1] = POP_BX;
			code [2] = PUSH;
			code [5] = PUSH;
			code [8] = PUSH_BX;
			code [9] = PUSH_AX;
			code [10] = FAR_JMP;
#else
			code [0] = POP_AX;
			code [1] = POP_BX;
			code [2] = PUSH;
			code [5] = PUSH_BX;
			code [6] = PUSH_AX;
			code [7] = FAR_JMP;
#endif
			hCode = (HGLOBAL)GlobalAlloc (GMEM_DDESHARE | GMEM_NOT_BANKED,
													InstanceCodeSize * NoOfInstances);
			if (hCode)
			{
				vfpCode = GlobalLock (hCode);
				if (vfpCode)
				{
					char far *fp = (char far *)vfpCode;
					for (int i = 0; i < NoOfInstances; i++,fp += InstanceCodeSize)
					{
						_fmemcpy (fp,&code,InstanceCodeSize);
						pFreeList->put ((FARPROC)fp);
					}
					PrestoChangoSelector (FP_SEG(vfpCode),FP_SEG(vfpCode));
				}
				else Status = FALSE;
			}
			else Status = FALSE;
		}
		else Status = FALSE;
	}
	else Status = FALSE;
}

/*

ObjectCallBack::~ObjectCallBack()

Function:

The class destructor frees the thunk global memory and Lists allocated in
the constructor.

*/

ObjectCallBack::~ObjectCallBack()
{
	if (hCode)
	{
		GlobalUnlock (hCode);
		GlobalFree (hCode);
	}
	if (pFreeList) delete pFreeList;
	if (pCallList) delete pCallList;
}

/*

FARPROC ObjectCallBack::MakeProcInstance (FARPROC fpPtr,PCObject pThis);

Function :

This function is the heart of the class; it fills in the this and call
back function pointers of the thunk and puts its address onto a Set.

AllocCStoDSAlias is used to create a data alias to the code selector. This
alias enables the pointers to be written without generating GP fails.

Returns:

Success - The thunk address.
Failure - NULL.

*/

FARPROC WINAPI ObjectCallBack::MakeProcInstance (FARPROC lpfPtr,PCObject pThis)
{
	if (Status)
	{
		char far *pf1 = (char far *)pFreeList->get ();
		if (!pf1) return NULL;
		char far *pfBlock = (char far *)MK_FP(AllocCStoDSAlias (FP_SEG(pf1)),FP_OFF(pf1));
	#if defined(__HUGE__) || defined(__LARGE__)
		*(unsigned int far *)(pfBlock + ThisSEG) = (unsigned int)FP_SEG(pThis);
		*(unsigned int far *)(pfBlock + ThisOFF) = (unsigned int)FP_OFF(pThis);
		*(unsigned long far *)(pfBlock + fpPtrSEG) = (unsigned long)lpfPtr;
	#else
		*(unsigned int far *)(pfBlock + ThisSEG) = (unsigned int)pThis;
		*(unsigned long far *)(pfBlock + fpPtrSEG) = (unsigned long)lpfPtr;
	#endif
		pCallList->add ((FARPROC)pf1);
		FreeSelector (FP_SEG(pfBlock));
		return (FARPROC)pf1;
	}
	else return NULL;
}

/*

static FARPROC ObjPtrToFarProc (CallObjTyp)

Function :

Converts an Object member function pointer to FARPROC.

Parameters:

ObjTyp - A pointer to the object member function.

Returns:

A pointer to the object member function as a FARPROC.

*/

#pragma warn -rvl
static FARPROC ObjPtrToFarProc (CallObjTyp)
{
#if defined(__HUGE__) || defined(__LARGE__)
	asm	mov	ax,[bp+0x06]
	asm	mov	dx,[bp+0x08]
#endif
#if defined(__SMALL__)
	asm	mov	ax,[bp+0x04]
	asm	mov	dx,[bp+0x06]
#endif
}
#pragma warn .rvl

/*

FARPROC WINAPI MakeProcInstance (CallObjTyp CallObjPtr,PCObject pThis)

Function :

Interface function passes a class member function and the this pointer
for encoding in the thunk and registering in the Instance array.

Parameters:

CallObjTyp CallObjPtr - A pointer to the object member function.
Note - The call back object member function cannot be virtual.

PCObject far *pThis - constant pointer to the instance of the object

Returns:

Success - A pointer to the thunk.
Failure - NULL

*/

FARPROC WINAPI ObjectCallBack::MakeProcInstance (CallObjTyp CallObjPtr,PCObject pThis)
{
	return (MakeProcInstance (ObjPtrToFarProc (CallObjPtr),pThis));
}

/*

FARPROC ObjectCallBack::FreeProcInstance (FARPROC lpfPtr)

Function :

This function searches the CallList for the thunk address. If the
address is found the CallList entry is removed, added to the FreeList
and NULL is returned. If no match is found nothing is done but lpfPtr
is returned.

Parameters:

FARPROC lpfPtr - A far pointer to the thunk address to be freed.

Returns:

Success - NULL
Failure - lpfPtr

*/

FARPROC WINAPI ObjectCallBack::FreeProcInstance (FARPROC lpfPtr)
{
	if (pCallList->hasMember (lpfPtr))
	{
		pCallList->detach (lpfPtr);
		pFreeList->put (lpfPtr);
		lpfPtr = NULL;
	}
	return lpfPtr;
}
